-- | This module contains functions for almost all the mathematical
--   concepts you will need to complete the assignment.
--   'Vector's and 'Colour's have been made instances of the
--   'Num' class so that you can add, subtract and multiply them where
--   this makes sense. For instance it does not make sense to multiply
--   'Vector's and this will yield an error. You can only peform
--   a dot or cross product upon them. 

--
-- YOU SHOULD NOT MODIFY ANYTHING IN THIS MODULE
--


module Physics
where

-- * New data types and type synonyms

-- | Describes only an object's shape, size and position.
data PureObject = Sphere Point    -- Centre
	                 Double   -- Radius
	        | Plane Distance  -- distance to origin
		        Vector    -- unit normal to surface 
     deriving (Show, Eq)

-- | An object with properties: diffuse reflection coefficient
--   specular reflection coefficient, and 'Colour'
data Object = Object PureObject --  Pure Object 
	             Double     --  diffuse reflection coefficient
                     Double     -- specular reflection coefficient
	             Colour     -- colour
               deriving (Show, Eq) 

-- Light. Has both a position (a 'Point') and a 'Colour'
data Light = Light  Point  -- position
	            Colour -- colour
     deriving  (Show, Eq)

-- | Points
newtype Point  = Pt  (Double, Double, Double) deriving (Show, Eq)

-- | Vectors
newtype Vector = Vec (Double, Double, Double) deriving (Show, Eq)

-- | Colours. Components should be in range [0.0, 1.0]
newtype Colour = Col (Double, Double, Double) deriving (Show, Eq)
                 

type Distance = Double
type Time = Double

-- | Rays have an origin (a 'Point') and a direction (a Vector). The direction
--   should be a unit vector.
data Ray = Ray Point   -- ray's origin
	       Vector  -- ray's direction (should be unit vector)
            deriving  (Show, Eq)


-- * Functions

-- ** Core functionality

-- | interesect determines whether a light ray and object intersect and
-- if so yields the time at which it intersects and a unit normal vector
-- at the point of intersection. Returns Nothing if no interesection occurs.
-- Note that intersection at negative times is possible.
intersect :: Ray -> PureObject -> Maybe (Time, Vector)
intersect ray (Plane d norm@(Vec (nx, ny, nz)))
  | discriminant == 0 = Nothing
  | otherwise = Just (t_int, Vec (s * nx, s*ny, s*nz))

  where Pt r0 = rayOrigin ray
        s = -signOf discriminant
        rd = rayDir ray
        discriminant = norm `dot` rd
        -- time value for interesection
        t_int = -(norm `dot` (Vec r0) + d) / discriminant

intersect ray (Sphere sc rad) 
     -- origin of light ray is in sphere
 | l_oc < rad = Nothing
 | t_ca > 0 &&  t_hc_sq < 0 = Nothing
 | t_ca < 0 = Nothing
 | otherwise = Just $ (t_isect, unit (vecFrom sc pt))
 where
  t_isect = t_ca - t_hc 
  pt = rayAt ray t_isect
  -- vector to the centre of the sphere from the origin of the ray
  r0 = rayOrigin ray
  rd = rayDir ray
  vectToCentre = vecFrom r0 sc 
  -- length from origin of vector to centre of sphere
  -- l_oc squared
  l_oc = magnitude vectToCentre
  t_ca = vectToCentre `dot` rd
  -- time it takes for ray to pass from intersection point to rayAt t_ca
  t_hc_sq = rad^2 - l_oc^2 + t_ca^2
  t_hc = sqrt t_hc_sq

-- | Checks whether a given 'Point' intersects with a pure object
intersectPt :: Point -> PureObject -> Bool
intersectPt pt (Sphere sc rad) = distBetween pt sc <= rad 
intersectPt (Pt (x,y,z)) (Plane d (Vec (nx, ny, nz))) = 
  nx*x + ny*y + nz*z ==* d 

-- | vecToLight takes a 'Point' and a light source and returns a 
--   unit vector pointing towards the light source
vecToLight :: Point -> Light -> Vector
vecToLight pt (Light pt' col) =
  let v = vecFrom pt pt' 
  in  unit v 


-- | Takes a 'Point' and a 'Light' and returns the intensity at that
--   point. Simulates light attenuation over distance.
intensity :: Point -> Light -> Colour
intensity pt (Light pt' col) =
  let v = vecFrom pt pt'
  in (100/ (99 + (v `dot` v))) `scaleCol` col

--
-- ** Ray utilities
--

-- | Takes a ray and a time parameter and yields the 'Point' on the ray
--   at that time.
rayAt :: Ray -> Time -> Point
rayAt ray t = 
  let (Pt (x0, y0, z0))  = rayOrigin ray  
      (Vec (xd, yd, zd)) = rayDir ray
  in  Pt (x0 + xd*t, y0 + yd*t, z0 + zd*t)

rayOrigin :: Ray -> Point 
rayOrigin (Ray orig _) = orig

rayDir :: Ray -> Vector
rayDir (Ray _ dir) = dir 

--
-- ** Vector utilities
--

-- | Determines the magnitude (or size) of a 'Vector'
magnitude :: Vector -> Distance
magnitude (Vec (x,y,z)) = sqrt (x^2 + y^2 + z^2)

-- | Takes two vectors and yields their dot product
dot :: Vector -> Vector -> Double
(Vec (x1, y1, z1)) `dot` (Vec (x2, y2, z2)) = x1*x2 + y1*y2 + z1*z2

-- | Takes two 'Point's and returns a 'Vector' pointing from the first
--   point to the second one. The result is not necessarily a unit vector.
vecFrom :: Point -> Point -> Vector
vecFrom (Pt (x1,y1,z1)) (Pt (x2,y2,z2)) = Vec (x2 - x1, y2 - y1, z2 - z1)


-- | Pre-multiplies a vector by a scalar
scaleVec :: Double -> Vector -> Vector
scaleVec k (Vec (ux, uy, uz)) = Vec (k*ux, k*uy, k*uz)

-- | Takes a vector and returns one that has the same direction but is a 
--   unit vector
unit :: Vector -> Vector
unit vec@(Vec (x,y,z)) = Vec (x/mag, y/mag, z/mag)
 where mag = magnitude vec

-- | Calculates the angle between two unit vectors
--   It will give the wrong answer for non-unit vectors
--   If you are uncertain, apply 'unit' to the arguments.
--   Angle is in radians between 0 and pi. This function will 
--   not tell you whether the second vector is clockwise or anti-clockwise
--   from the first.
angleBetween :: Vector -> Vector -> Double
angleBetween v1 v2 = acos (v1 `dot` v2 )
  
-- | The distance between two points
distBetween :: Point -> Point -> Double
distBetween p1 p2 = magnitude (vecFrom p1 p2)

-- | Rotates a vector by theta radians in the plane defined
--   by the normal vector that is the first argument.
--   The rotation is clockwise when one views the plane with the 
--   unit normal vector pointing towards oneself.  Equivalently, the rotation
--   is anti-clockwise when viewing the plane with the normal vector
--   pointing away from oneself.  Negative arguments mean an
--   anti-clockwise rotation.  This function preserves the magnitude
--   of the vector. If the original vector was a unit vector, so is
--   the result.
--   Precondition: 
--      1. Expects that the normal vector is a unit normal vector
--      2. The vector to be rotated must be in the plane. i.e.
--         It must be perpendicular to the unit normal vector

rotateInPlane :: Vector -- ^ unit normal vector
	      -> Vector -- ^ vector
	      -> Double -- ^ theta
	      -> Vector
rotateInPlane (Vec (nx, ny, nz)) (Vec (ax, ay, az)) theta =
  Vec ( az*ny*sin_theta - ay*nz*sin_theta + ax*cos_theta
      , ax*nz*sin_theta - az*nx*sin_theta + ay*cos_theta
      , ay*nx*sin_theta - ax*ny*sin_theta + az*cos_theta)
 where sin_theta = sin theta
       cos_theta = cos theta

-- | Cross product of a 'Vector'. Yields a 'Vector' that is 
--   is perpendicular to both arguments. Does not necessarily yield
--   a unit vector as a result.
cross :: Vector -> Vector -> Vector
cross (Vec (ax,ay,az)) (Vec (bx,by,bz)) = 
  Vec (ay*bz-az*by, az*bx-ax*bz, ax*by - ay*bx)



--
-- ** Colours 
--

-- | Pre-multiplies a 'Colour' by a scalar
scaleCol :: Double -> Colour -> Colour
k `scaleCol` (Col (r,g,b)) = (Col (k*r, k*g, k*b))

-- *** Sample colours

-- | Brightest red
red :: Colour
red = Col (1,0,0)

-- | Brightest green
green :: Colour
green = Col (0,1,0)

-- | Brightest blue
blue :: Colour
blue = Col (0,0,1)

-- | Pure black
black :: Colour
black = Col (0,0,0)

-- | Brightest white
white :: Colour
white = Col (1,1,1)

-- | signOf is a lot like signum except it returns 1 for 0.0.
--   This function returns the sign of a floating-point number.
signOf x = if x >= 0 then 1 else -1

--------------------------

-- ** Miscellaneous functions

-- | The resolution for doubles. If two doubles differ by less than this
--   amount we consider them equal
doubleRes :: Double
doubleRes = 0.0001

infix 4 ==* 

-- | Equality up to a given resolution (equal to 'doubleRes')
(==*) :: Double -> Double -> Bool
a ==* b = abs (a - b) < doubleRes


-- | Takes an object and returns the pure object contained within it.
getPureObj :: Object -> PureObject
getPureObj (Object pureObj _ _ _) = pureObj


instance Num Vector where
  (Vec (ux,uy,uz)) + (Vec (vx,vy,vz)) = Vec (ux + vx, uy + vy, uz + vz)
  (Vec (ux,uy,uz)) * (Vec (vx,vy,vz)) = 
    error "Multiplication is undefined for Vector. Use 'dot' or 'cross'"
  (Vec (ux,uy,uz)) - (Vec (vx,vy,vz)) = Vec (ux - vx, uy - vy, uz - vz)
  negate (Vec (ux,uy,uz)) = Vec (-ux, -uy, -uz)
  abs _ = error "abs not defined for Vector"
  signum _ = error "signum not defined for Vector"
  fromInteger _ = error "fromInteger not defined for Vector"

instance Num Colour where
  (Col (r1, g1, b1)) + (Col (r2, g2, b2)) = Col (r1 + r2, g1 + g2, b1 + b2)
  (Col (r1, g1, b1)) * (Col (r2, g2, b2)) = Col (r1 * r2, g1 * g2, b1 * b2)
  (Col (r1, g1, b1)) - (Col (r2, g2, b2)) = Col (r1 - r2, g1 - g2, b1 - b2)
  negate _ = error "negate not defined for Colour"
  abs _ = error "abs not defined for Colour"
  signum _ = error "signum not defined for Colour"
  fromInteger _ = error "fromInteger not defined for Colour"