module RayTrace
where

-- Library imports
import Data.List hiding (intersect)
import Data.Maybe (isJust, fromJust)

-- local imports
import Physics
import World


-- Takes an OctTree representing the scene, a list of light sources, a
-- ray from the eye position and calculates the colour the pixel
-- should be.
-- Precondition: rayAt ray 0 must be within the octTree.
rayTrace :: Int -> Scene -> Ray ->  Colour
rayTrace 0 _ _   = black
rayTrace depth scene@(Scene _ minSize lights ambient) ray 
  | isJust isectInfo  = 
     -- intersection found!
     ambient + diffuse + specular + otherObjects
  | otherwise = black
 where
  isectInfo = firstIntersect scene (getLeafContaining 
				    (rayAt ray 0) scene) ray  

  (obj, t, norm) = fromJust isectInfo
  (Object _ k_d k_s col) = obj
  ambient = ambientComponent scene ray (obj, t, norm)
  diffuse = diffuseComponent scene ray (obj, t, norm)
  specular = specularComponent scene ray (obj, t, norm)
  otherObjects = otherObjectsComponent depth scene ray (obj, t, norm)


-- Calculates the ray from the eyePosition for the given column i and row j.
rayForPixel :: Double -> Double -> (Double, Double) -> (Int,Int) -> Ray
rayForPixel wldWidth pixelWidth (x,y) (i,j) = 
   Ray eyePosition dir
 where 
  i' = fromIntegral i
  j' = fromIntegral j
  dir = unit (Vec (x + (i' + 0.5)*pixelWidth, y - (j' + 0.5)*pixelWidth, 1))

-- Calculates the ambient component of the light entering the eye.
-- The second argument is the ray from the eye.
-- The third argument consists of the object the ray from the eye
-- has intersected with, the time at which that occurs and the unit normal
-- vector at that point on the surface.
ambientComponent :: Scene -> Ray -> (Object, Time, Vector) -> Colour
ambientComponent (Scene _ _ ambient _) ray (obj, t, norm) = 
  k_d `scaleCol` (ambient * col)
  where
    Object _ k_d _ col = obj

-- Takes a list of objects, a voxel and a ray and yields a (possibly
-- empty) list of ordered triples.  Each triple contains an object, the time
-- at which the ray intersected the object and a unit normal vector at the 
-- point of intersection.
-- The list is ordered such that the interesection with the smallest
-- time value is first, followed by the next greatest time value and so on.
intersectsInVoxel :: [Object] -> Voxel -> Ray -> [(Object, Time, Vector)]
intersectsInVoxel = 
  error $ 
    "RayTrace.intersectsInVoxel is undefined.\n" ++
    "*** The implementation of this functions is part of the BONUS ***\n" ++
    "If you want to test a non-bonus solution, download the binary file\n" ++
    "of RayTrace from the assignment web page."


-- Takes a scene, a leaf, and a ray (which passes through the leaf).
-- Finds the first (non-zero) intersection with an object, if one
-- exists, and returns the object and the time at which the ray
-- intersects the object. This function will check if there are any
-- intersections in the leaf.  If not it will try the next leaf that
-- ray passes through. This process terminates when an intersection is
-- found or the ray passes out of the scene (i.e. leaves the voxel
-- containing the entire scene) If no intersection exists Nothing is
-- returned. This function should not return a time value of zero
-- Precondition : Ray passes through leaf
firstIntersect :: Scene -> OctTree -> Ray ->  Maybe (Object, Time, Vector)
firstIntersect = 
  error $
    "RayTrace.firstIntersect is undefined.\n" ++
    "*** The implementation of this functions is part of the BONUS ***\n" ++
    "If you want to test a non-bonus solution, download the binary file\n" ++
    "of RayTrace from the assignment web page."


-- Checks whether there is an intersection with another object
-- before the light source. Takes a scene, the point of intersection,
-- a light source, and a ray to that light as arguments.
-- Precondition: The ray must actually point at the light.
checkShadowIntersect :: Scene -> Point -> Light -> Ray -> Bool
checkShadowIntersect = 
  error $ "RayTrace.checkShadowIntersect is undefined.\n" ++
    "*** The implementation of this functions is part of the BONUS ***\n" ++
    "If you want to test a non-bonus solution, download the binary file\n" ++
    "of RayTrace from the assignment web page."


-- getLightIntensity takes a scene, a point of intersection with an
-- object, the unit normal vector at that point on the surface and a
-- pair consisting of a light and a vector pointing from the point to
-- the light source. It then calculates what the intensity of that
-- light source should be at that point. This function creates a
-- "shadow" ray and checks for intersection with other objects. If
-- such an intersection occurs the light intensity is nothing
-- (i.e. black)
getLightIntensity :: Scene -> Point -> Vector 
		  -> (Light, Vector) -> Colour
getLightIntensity scene isectPt norm (light, lightVec) = 
 if angle > pi/2 then black else
  if checkShadowIntersect scene isectPt light rayToLight 
   then black else intensity isectPt light
 where
    rayToLight = Ray isectPt lightVec
    angle = angleBetween norm lightVec

-- Calculates the diffuse component of the light entering the eye.
-- The second argument is the ray from the eye.
-- The third argument consists of the object the ray from the eye
-- has intersected with, the time at which that occurs and the unit normal
-- vector at that point on the surface.
diffuseComponent :: Scene -> Ray -> (Object, Time, Vector) -> Colour
diffuseComponent scene@(Scene _ _ _ lights) ray (obj, t, norm) = 
  foldr (+) black (zipWith3 dotMult normals lightVecs intensities)
 where
  Object _ k_d _ col = obj
  isectPt = rayAt ray t
  normals = repeat norm
  lightsAndVecs = zip lights lightVecs
  lightVecs = map (vecToLight isectPt) lights
  intensities = map (getLightIntensity scene isectPt norm)
                    lightsAndVecs
  dotMult n l_j i_j = ((k_d * (n `dot` l_j)) `scaleCol` i_j) * col

-- Calculates the specular component of the light entering the eye.
-- The second argument is the ray from the eye.
-- The third argument consists of the object the ray from the eye
-- has intersected with, the time at which that occurs and the unit normal
-- vector at that point on the surface.
specularComponent :: Scene -> Ray -> (Object, Time, Vector) -> Colour
specularComponent scene@(Scene _ _ _ lights) ray (obj, t, norm) =
  foldr (+) black (zipWith3 dotMult normals halfWayVects intensities)
 where

  Object _ _ k_s col = obj
  isectPt = rayAt ray t
  normals = repeat norm
  dotMult n h_j i_j = ((k_s * (n `dot` h_j)) `scaleCol` i_j) * col
  -- unit normal of plane containing the object normal and the ray
  opRayDir = negate (rayDir ray) -- opposite direction of ray

  halfWayVect lightVec = 
    let plane = unit (cross lightVec opRayDir)
    in  rotateInPlane plane lightVec ((angleBetween lightVec opRayDir)/2)
  halfWayVects = map halfWayVect lightVecs

  lightsAndVecs = zip lights lightVecs
  lightVecs = map (vecToLight isectPt) lights
  intensities = map (getLightIntensity scene isectPt norm) 
		    lightsAndVecs

-- Calculates the component of the light entering the eye due to reflections
-- from other objects. 
-- The first argument is the _current_ depth of ray tracing recursion.
-- (It is _this_ function that decrements the recursion depth before
--  recursively calling rayTrace.)
-- The third argument is the ray from the eye.
-- The fourth argument consists of the object the ray from the eye
-- has intersected with, the time at which that occurs and the unit normal
-- vector at that point on the surface.
otherObjectsComponent :: Int -> Scene -> Ray 
		      -> (Object, Time, Vector) -> Colour
otherObjectsComponent depth scene ray@(Ray orig rayDir) (obj, t, norm) =
  let newray = Ray isectPt reflectVec
  in  (k_s `scaleCol` i_s) * col
  
 where 
  i_s = rayTrace (depth-1) scene (Ray isectPt reflectVec)
  Object _ _ k_s col = obj
  isectPt = rayAt ray t 
  reflectVec =  let plane = unit (cross opRayDir norm)
                in  rotateInPlane plane norm (angleBetween opRayDir norm)
  opRayDir = negate rayDir
