--
-- module Main
-- 
-- YOU SHOULD NOT MODIFY ANYTHING IN THIS MODULE
--
--

module Main where

-- Library imports
import System.IO
import System.Environment
import System.Exit
import Data.Maybe

-- local imports
import Physics
import SceneParser
import PpmWriter
import World
import RayTrace

-- The main function. Checks whether a scene file has been provided on the
-- command line, then tries to parse the scene file. If this succeeds a
-- check is performed on light sources and the eye position to see if they
-- overlap with any objects. If they don't then the scene is rendered via
-- a call to render. Any errors in command line arguments, scene files or 
-- failure of the overlap check are reported to the user on stdout.
main :: IO ()
main = do
  -- Turn of stdout buffering
  hSetBuffering stdout NoBuffering

  progName <- getProgName
  args <- getArgs 
  if length args < 1 
   then do putStrLn ("Usage: " ++ progName ++ " <scene description file>")
           exitWith (ExitFailure 1)
   else return ()
  let sceneFile = head args
  sceneStr <- readFile sceneFile
  let mbSceneInfo = parseScene sceneStr
  if isError  mbSceneInfo 
   then do let Error errorStrings = mbSceneInfo
           putStr (unlines (filter (\s -> (not (null s))) errorStrings))
	   exitWith (ExitFailure 1)
   else return ()
  let None (filePath, depth, fov, size@(w,h)
	   , ambient, objs, lights) = mbSceneInfo
  let mbLightsOK = checkScene objs lights
  if isJust mbLightsOK 
   then do putStrLn ("Scene description is syntactically valid but:\n  " ++ 
		     fromJust mbLightsOK)

           exitWith (ExitFailure 1)
   else return ()
  -- Lights okay
  let aspectRatio = fromIntegral w / fromIntegral h
      boundVox = boundingVoxel fov aspectRatio 
      scene = mkScene ambient objs lights boundVox 
  render scene depth fov size filePath

-- Given a scene, a depth of recursion, a field of view, a size (of the final
-- image and the path of a file, renders the scene and writes the
-- image to the file.
render :: Scene ->  Int -> Double -> Size -> FilePath  -> IO ()
render scene depth fov (wd,ht) filePath = do
     putStrLn $ "width=" ++ show wldWidth ++ " height=" ++ show wldHeight
     writePpm filePath 
	      ((wd, ht), 
	       [[pixelFun x y | y <- [1..ht]] | x <- [1..wd]])
 where 
  pixelFun i j = 
    let ray = rayForPixel wldWidth pixelWidth (-wldWidth/2,wldHeight/2) (i,j)
    in  rayTrace depth scene ray
  minSize = getMinSize scene
  wldWidth = 2 * tan (0.5 * fov * pi/180)
  wldHeight = wldWidth * (fromIntegral ht) / (fromIntegral wd)
  -- The width of a pixel
  pixelWidth = wldWidth / fromIntegral wd
