--
-- module SceneParser
--
-- YOU SHOULD NOT MODIFY ANYTHING IN THIS MODULE
--


module SceneParser 
where

import Physics
import World


import Data.Char
import Data.Either
import Debug.Trace

type SceneInfo = ( FilePath
		 , Int
		 , Double
		 , (Int, Int)
		 , Colour
		 , [Object]
		 , [Light]
		 )

type Item = Either Object Light

data Error a = None a 
	     | Error [String]

----------------------------------------

isNone :: Error a -> Bool
isNone (None _) = True
isNone _        = False

isError :: Error a -> Bool
isError (Error _) = True
isError _         = False

fromNone :: Error a -> a
fromNone (None a) = a
fromNone _ = error "fromNone called on Error"


errorStrs (Error strs) = strs
errorStrs (None _) = []

condString :: Bool -> String -> String
condString b str = if b then str else ""


----------------------------------------

headerLength :: Int
headerLength = 5

parseScene :: String -> Error SceneInfo
parseScene sceneStr 
 | length sceneLines >= headerLength && isNone mbHeader 
   && all isNone mbItems = 
     let (filePath, depth, fov, size, ambient) = fromNone  mbHeader
     in None  (filePath, depth, fov, size, ambient, objs, lights)
 | otherwise = Error (errorStrs  mbHeader ++ (concat (map errorStrs mbItems)))

 where
  sceneLines = map words (lines (map toLower sceneStr))
  headerLines = take headerLength sceneLines
  restLines = drop headerLength sceneLines
  mbItems = map parseItem restLines
  items = map fromNone mbItems
  getItems [] = ([], [])
  getItems (item:items) = 
    let (objs, lights) = getItems items
    in  case item of 
         Left   obj   -> (obj:objs, lights)
         Right  light -> (objs, light:lights)
  (objs, lights) = getItems items
  mbHeader = parseHeader headerLines

parseHeader :: [[String]] -> Error (FilePath, Int, Double, (Int, Int), Colour)
parseHeader [ ["file", filePath] 
	    , ["depth", depthStr]
	    , ["fov", fovStr]
	    , ["size", wStr, hStr]
            , ["ambient", rStr, gStr, bStr]
	    ]

 | isNone mbColour && isNone mbFov && isNone mbW && isNone mbH && 
   isNone mbDepth && 1 <= fov && fov <= 179 = 

     None  (filePath, fromNone mbDepth, fov
	   , (fromNone mbW, fromNone mbH), fromNone  mbColour)
 | otherwise = Error (concat [ errorStrs mbColour
			     , errorStrs mbW
			     , errorStrs mbH
			     , errorStrs mbDepth
			     , errorStrs mbFov
			     ] ++ 
		      [ condString (fov < 1 || fov > 179) 
			"Field of view, which is " ++ show fov ++
			", must be between 1.0 and 179.0"
		      ])

 where 
  mbColour = mkColour rStr gStr bStr 
  mbFov = readDouble fovStr
  fov = fromNone  mbFov
  mbW = readSize wStr
  mbH = readSize hStr
  mbDepth = readSize depthStr


parseHeader header = Error ["Header of this scene file is not valid."]



parseItem :: [String] -> Error Item
parseItem ["sphere", xStr, yStr, zStr, radStr
	  , diffStr, specStr, rStr, gStr, bStr]
 | isNone mbRad && isNone mbDiff && isNone mbSpec && isNone mbColour &&  
   isNone mbPoint && rad > 0 =
   None  $ Left (Object (Sphere (fromNone  mbPoint) rad) 
		  (fromNone  mbDiff) (fromNone  mbSpec) (fromNone  mbColour))
 | otherwise = Error (concat [ errorStrs mbRad
			     , errorStrs mbDiff
			     , errorStrs mbSpec
			     , errorStrs mbColour
			     , errorStrs mbPoint] ++
		      [ condString (rad <= 0) 
			("Sphere radius " ++ show rad ++ 
			 " is not postive.")]
		     )
 where mbRad = readDouble radStr
       mbDiff = readCoeff diffStr
       mbSpec = readCoeff specStr
       mbColour = mkColour rStr gStr bStr
       mbPoint = mkPoint xStr yStr zStr
       rad = fromNone  mbRad

parseItem ["plane", distStr, vxStr,vyStr,vzStr, diffStr, specStr
	  , rStr, gStr, bStr]
 | isNone mbColour && isNone mbUnitVec && isNone mbDiff &&
   isNone mbSpec  && isNone mbDist && dist >= 0 = 
   None  $ Left (Object (Plane dist (fromNone  mbUnitVec))
                       (fromNone  mbDiff) (fromNone  mbSpec) 
		       (fromNone  mbColour))
 | otherwise = Error (concat [ errorStrs mbColour
			     , errorStrs mbUnitVec
			     , errorStrs mbDiff
			     , errorStrs mbSpec
			     , errorStrs mbDist 
			     ] ++
		      [ condString (dist < 0) 
			("Plane distance " ++ show dist ++ 
			 " is less than zero")
		      ])
 where mbColour = mkColour rStr gStr bStr
       mbUnitVec = mkUnitVec vxStr vyStr vzStr
       mbDiff = readCoeff diffStr
       mbSpec = readCoeff specStr 
       mbDist = readDouble distStr
       dist = fromNone  mbDist

parseItem ["light", xStr, yStr, zStr, rStr, gStr, bStr]
 | isNone mbPoint && isNone mbColour =
   None  $ Right (Light (fromNone  mbPoint) (fromNone  mbColour))
 | otherwise = Error (concat [errorStrs mbPoint, errorStrs mbColour])
 where mbPoint = mkPoint xStr yStr zStr
       mbColour = mkColour rStr gStr bStr

parseItem _ = Error ["Valid light, sphere or plane required on each" ++
		     "line after header."]

readCoeff :: String -> Error Double
readCoeff str 
  | isNone mbVal && inRange val = None  (fromNone  mbVal)
  | otherwise = 
     Error (errorStrs mbVal  ++ 
	    if isNone mbVal
	       then ["Coefficient " ++ show val ++ " not " ++
		   "between 0.0 and 1.0."]
            else errorStrs mbVal
	   )
 where mbVal = readDouble str
       val = fromNone mbVal
 
mkUnitVec :: String -> String -> String -> Error Vector
mkUnitVec vxStr vyStr vzStr = 
  mkGenericTriple vxStr vyStr vzStr 
                  (unit . Vec) 
		  (\str -> "You'll never see me as an error.")
		  (const True)

mkPoint :: String -> String -> String -> Error Point
mkPoint xStr yStr zStr = 
  mkGenericTriple xStr yStr zStr 
                  Pt 
		  (\str -> "You'll never see me as an error.")
		  (const True)

mkColour :: String -> String -> String -> Error Colour
mkColour rStr gStr bStr = 
  mkGenericTriple rStr gStr bStr 
                  Col 
		  (\dbl -> show dbl ++ " is not a valid colour component." ++
		   " Must be in range 0.0 to 1.0.")
		  inRange

mkGenericTriple :: String -> String -> String 
		-> ((Double, Double, Double) -> a) 
		-> (Double -> String)
		-> (Double -> Bool)
		-> Error a
mkGenericTriple aStr bStr cStr maker errFn checker
 | all isNone mbDoubles = 
   let [a,b,c] = map fromNone  mbDoubles
       ok = checker a && checker b && checker c 
   in if ok then None  (maker (a,b,c)) 
            else Error (map errFn 
			(filter (\x -> not (checker x)) [a,b,c]))
 | otherwise = Error (concat (map errorStrs mbDoubles))
 where  mbDoubles = map readDouble [aStr,bStr,cStr]

inRange :: Double -> Bool
inRange x = 0.0 <= x && x <= 1.0

readDouble :: String -> Error Double
readDouble str
 | isDoubleStr str = None (read str)
 | otherwise = Error $ [str ++ " is not a valid Double."]


-- Checks whether string is a valid double of the form
-- [-]X or [-]X.X where X stands for any string of digits
-- and the [-] is an optional negative sign.
isDoubleStr :: String -> Bool
isDoubleStr "" = False
isDoubleStr str = 
  not (null numPart) && 
  ((all isDigit numPart)  -- literal of form X
  || 
  ( middleChar == '.' && (length minus <= 1) && 
    not (null front)  && not (null back) && 
    all isDigit back)) -- literal of form X.X
 where 
   minus = takeWhile (=='-') str
   numPart  = dropWhile (=='-') str
   front  = takeWhile isDigit numPart
   (middleChar:back) = dropWhile isDigit numPart

readSize :: String -> Error Int
readSize str = if isSizeStr str 
                then if sz > 0 
                      then None sz
                      else Error [show sz ++ " is not positive."]
                else Error [str ++ " is not a valid size."]
        
 where 
   isSizeStr str = all isDigit str -- negative numbers not allowed!
   sz = read str