--
-- module WorldBase
--
-- YOU SHOULD NOT MODIFY ANYTHING IN THIS MODULE
--

module WorldBase
where

-- Library imports
import Data.List hiding (intersect)
import Data.Maybe


-- local imports
import Physics



----------------------------------------------------------------------

data OctTree =  Leaf Voxel [Object] -- objects contained within this node
             |  Node Voxel [OctTree]
                 deriving Show

data Scene = Scene OctTree  -- octTree containing the scene
                   Distance -- minsize
                   Colour   -- ambient lighting
                   [Light]  -- lights in the scene 
                  deriving Show

type Size = ( Int, Int) -- size of the image in pixels. (width, height)

type Voxel = ( Point   -- bottom/left/front corner
             , Distance) -- length of side 


----------------------------------------------------------------------


-- The position of the eye
eyePosition = Pt (0,0,-1) 

maxObjects :: Int
maxObjects = 5 -- the maximum number of objects in a voxel


-- Gets the voxel containing the whole scene
sceneVoxel :: Scene -> Voxel
sceneVoxel (Scene (Leaf voxel _) _ _ _ ) = voxel
sceneVoxel (Scene (Node voxel _) _ _ _ ) = voxel

voxelSize :: Voxel -> Distance
voxelSize (_, size) = size

getMinSize :: Scene -> Distance
getMinSize (Scene _ minSize _ _) = minSize



isInVoxel :: Point -> Voxel -> Bool
isInVoxel (Pt (x,y,z)) (Pt (x0,y0,z0), size) =
  x0 <= x && x < x0 + size &&
  y0 <= y && y < y0 + size &&
  z0 <= z && z < z0 + size 

contains :: Voxel -> PureObject -> Bool
-- Sphere/voxel intersection algorithm taken from p599 of 
-- T. Akenine-Moeller and Eric Haines. Real-time Rendering, Second
-- Edition. Algorithm originally due to James Arvo. "A Simple Method
-- for Box-Sphere Intersection Testing" in Graphics Gems. pp335-339
-- 
contains vox@(p1@(Pt (x0,y0,z0)), size) (Sphere c@(Pt (xc,yc,zc)) r)  = 
  not voxelInside && distSquared < r_sq
 where
   r_sq = r^2
   p2@(Pt (x1,y1,z1)) = getTopCorner vox
   distSquared = distToNearestFace xc x0 x1 +
		 distToNearestFace yc y0 y1 +
   		 distToNearestFace zc z0 z1  
   -- Takes a point and two bounding points. If the first point is between
   -- the bounding points zero is returned. Otherwise the distance
   -- squared to the nearest of the bounding points is returned.
   distToNearestFace a a_min a_max =
     if a < a_min then (a - a_min)^2 else 
      if a >= a_max then (a - a_max)^2 else 0
   -- checks to see if sphere completely contains the box.
   voxelInside = 
     let v1 = vecFrom c p1 
         v2 = vecFrom c p2
     in  (v1 `dot` v1) <= r_sq && (v2 `dot` v2) < r_sq

-- Plane/voxel intersection algorithm taken from p587 of 
-- T. Akenine-Moeller and Eric Haines. Real-time Rendering, Second
-- Edition.
contains ((Pt (x0,y0,z0)), size) (Plane d norm@(Vec (nx, ny, nz))) =
  norm `dot` v_min + d < 0 && norm `dot` v_max + d > 0
 where   
   v_min = Vec (vx_min, vy_min, vz_min)
   v_max = Vec (vx_max, vy_max, vz_max)
   (vx_min, vx_max) = diagTest nx x0 (x0 + size)
   (vy_min, vy_max) = diagTest ny y0 (y0 + size)
   (vz_min, vz_max) = diagTest nz z0 (z0 + size)
   diagTest na a0 a1 = if na >= 0 then (a0,a1) else (a1, a0)


-- If we fire a ray through a voxel it will exit and some point and enter
-- another. This function yields a point guaranteed to be in the next
-- voxel. Requires the minimum voxel length of the scene as a parameter.
pointNextVoxel ::  Distance -> Ray -> Voxel -> Point
pointNextVoxel minLen ray@(Ray orig dir@(Vec (dx, dy, dz))) 
                      vox@(Pt (x0,y0,z0), size) 
  = Pt (x_base + x_delta, y_base + y_delta, z_base + z_delta)
 where
   Pt (x_base, y_base, z_base) = rayAt ray t_plus 
   Pt (x1, y1, z1) = getTopCorner vox
   t_plus = minimum t_list
   x_delta = deltaCalc mbi_x0 mbi_x1 (signOf dx)
   y_delta = deltaCalc mbi_y0 mbi_y1 (signOf dy) 
   z_delta = deltaCalc mbi_z0 mbi_z1 (signOf dz)
   -- Works out whether a point should be shifted on a particular axis
   deltaCalc mt1 mt2 sig = 
      let delta = sig*minLen/2
      in case mt1 of 
          Just (t, _) -> 
           (if t ==* t_plus  then delta else 0) +
           (let (t', _) = fromJust mt2 
            in if t'==* t_plus then delta else 0)
          Nothing -> 0
   -- Makes a plane out of a face co-ordinate
   mkPlane a dir = Plane (abs a) ((-signOf a) `scaleVec` dir)
   --
   maxOfAxisPair mb_isect mb_isect' = 
     case mb_isect of 
      Just (t,_) -> let  (t', _) = fromJust mb_isect'
                    in [max t t']
      Nothing -> [] 
   t_list = t_x ++ t_y ++ t_z
   t_x = maxOfAxisPair mbi_x0 mbi_x1
   t_y = maxOfAxisPair mbi_y0 mbi_y1
   t_z = maxOfAxisPair mbi_z0 mbi_z1
   -- 
   mbi_x0 = intersect ray (mkPlane x0 (Vec (1,0,0)))
   mbi_x1 = intersect ray (mkPlane x1 (Vec (1,0,0)))
   mbi_y0 = intersect ray (mkPlane y0 (Vec (0,1,0)))
   mbi_y1 = intersect ray (mkPlane y1 (Vec (0,1,0)))
   mbi_z0 = intersect ray (mkPlane z0 (Vec (0,0,1)))
   mbi_z1 = intersect ray (mkPlane z1 (Vec (0,0,1)))

-- Given a voxel yields the bottom/left/front corner
getBottomCorner :: Voxel -> Point
getBottomCorner (pt, _) = pt

-- Given a voxel yields the top/right/back corner.
getTopCorner :: Voxel -> Point
getTopCorner ((Pt (x0,y0,z0)), size) = Pt (x0 + size, y0 + size, z0 + size)


-- checkScene checks whether any of the light sources intersect with
-- an objects. Returns Nothing if everything is okay, otherwise it returns
-- a string with details of the problem.
-- Also checks whether the eyePosition is inside an object
checkScene :: [Object] -> [Light] -> Maybe String
checkScene objs lights = 
  case checkLights objs lights of 
    Just str -> Just str
    Nothing -> checkEyePos objs 
 where 
  checkLights [] lights = Nothing 
  checkLights (obj:objs) lights
    | all isNothing isects = checkLights objs lights
    | otherwise = Just (show obj ++ " intersects with " ++ show fstLight)
   where 
     Object pureObj _ _ _ = obj
     isects = map getIntersect lights
     getIntersect light@(Light pt _) = 
       if intersectPt pt pureObj then Just light else Nothing
     fstLight = head (dropWhile isNothing isects)
  checkEyePos [] = Nothing
  checkEyePos (obj@(Object pureObj _ _ _):objs) 
    | intersectPt eyePosition pureObj = 
      Just (show obj ++ " interesects with eye position at "  
	    ++ show eyePosition)
    | otherwise = checkEyePos objs



boundMultiplier :: Double
boundMultiplier = 100

-- Works out the voxel that is to bound the scene. It is
-- 'boundMultiplier' times the size of the maximum of the width and
-- height of the view plane. The view plane is calculated from the
-- field of view (fov) and the aspect ratio 
-- Precondition: 1 <= fov <= 179. aspectRatio > 0 
boundingVoxel :: Double -> Double -> Voxel
boundingVoxel fov aspectRatio = (Pt (-size/2, -size/2, -2), size)
 where 
  width = 2*tan(0.5 * fov * pi/180)
  height = width / aspectRatio
  size  = boundMultiplier * max width height

----------------------------------------
-- Helper functions

sizeOctTree :: OctTree -> Int
sizeOctTree (Leaf _ _) = 1
sizeOctTree (Node _ octTrees) = 1 + sum (map sizeOctTree octTrees)

sizeScene :: Scene -> Int
sizeScene (Scene oct _ _ _ ) = sizeOctTree oct